# this class runs the "AgentProcess" for the agents and holds the pool
# of agents currently usable
# this class isn't actually instantiated until the user instantiates
# a specific agent like MathAgent

import heapq
from threading import Lock, Event
from pympler import asizeof
from .interact import Interactor
import os
import importlib

class AgentFactory:
    """ duplicate of AgentProcessFactory """

    def __init__(self, llm, agent_process_queue, agent_process_factory, agent_log_mode):
        # 256 agent ids heapified similar to AgentProcessFactory
        self.max_aid = 256
        # self.llm = llm
        self.aid_pool = [i for i in range(self.max_aid)]
        heapq.heapify(self.aid_pool)
        self.agent_process_queue = agent_process_queue

        self.agent_process_factory = agent_process_factory


        # added to with index aid
        self.current_agents = {}

        self.current_agents_lock = Lock()

        self.terminate_signal = Event()

        self.agent_log_mode = agent_log_mode

    def snake_to_camel(self, snake_str):
        components = snake_str.split('_')
        return ''.join(x.title() for x in components)

    def list_agents(self):
        agent_list = Interactor().list_available_agents()
        for agent in agent_list:
            print(agent)

    def load_agent_instance(self, agent_name):
        author, name = agent_name.split("/")
        module_name = ".".join(["pyopenagi", "agents", author, name, "agent"])
        class_name = self.snake_to_camel(name)

        agent_module = importlib.import_module(module_name)

        agent_class = getattr(agent_module, class_name)
        return agent_class

    def activate_agent(self, agent_name, task_input):
        script_path = os.path.abspath(__file__)
        script_dir = os.path.dirname(script_path)

        interactor = Interactor()

        if not os.path.exists(os.path.join(script_dir, agent_name)):
            interactor.download_agent(agent_name)

        if not interactor.check_reqs_installed(agent_name):
            interactor.install_agent_reqs(agent_name)

        agent_class = self.load_agent_instance(agent_name)

        """ initialize each agent """
        agent = agent_class(
            agent_name = agent_name,
            task_input = task_input,
            agent_process_factory = self.agent_process_factory,
            log_mode = self.agent_log_mode
        )

        aid = heapq.heappop(self.aid_pool)

        agent.set_aid(aid)

        if not self.terminate_signal.is_set():
            with self.current_agents_lock:
                self.current_agents[aid] = agent

        return agent

    def run_agent(self, agent_name, task_input):
        """ run the Thread and return output """

        agent = self.activate_agent(
            agent_name=agent_name,
            task_input=task_input
        )
        # print(task_input)
        output = agent.run()
        self.deactivate_agent(agent.get_aid())
        return output

    def print_agent(self):
        """ print all agent information in a nice organized format """
        headers = ["Agent ID", "Agent Name", "Created Time", "Status", "Memory Usage"]
        data = []
        for id, agent in self.current_agents.items():
            agent_name = agent.agent_name
            created_time = agent.created_time
            status = agent.status
            memory_usage = f"{asizeof.asizeof(agent)} bytes"
            data.append(
                [id, agent_name, created_time, status, memory_usage]
            )
        self.print(headers=headers, data=data)


    def print(self, headers, data):
        """ generalized table printer for the agent data """
        # align output
        column_widths = [
            max(len(str(row[i])) for row in [headers] + data) for i in range(len(headers))
        ]
        print("+" + "-" * (sum(column_widths) + len(headers) * 3 - 3 ) + "+")
        print(self.format_row(headers, column_widths))
        print("=" * (sum(column_widths) + len(headers) * 3 - 1))
        for i, row in enumerate(data):
            print(self.format_row(row, column_widths))
            if i < len(data):
                print("-" * (sum(column_widths) + len(headers) * 3 - 1))
        print("+" + "-" * (sum(column_widths) + len(headers) * 3 - 3 ) + "+")


    def format_row(self, row, widths, align="<"):
        """ helper utility for print """
        row_str = " | ".join(f"{str(item):{align}{widths[i]}}" for i, item in enumerate(row))
        return row_str

    def deactivate_agent(self, aid):
        """ remove the agent id from the pool and allow it to be re-used """

        self.current_agents.pop(aid)
        heapq.heappush(self.aid_pool, aid)



class BaseTool:
    """Base class for calling tools
    """
    def __init__(self) -> None:
        pass

    def run(self, params) -> None:
        pass

    def get_tool_call_format(self) -> dict:
        """Get tool calling format following the function calling from OpenAI: https://platform.openai.com/docs/guides/function-calling
        """
        pass

class BaseRapidAPITool(BaseTool):
    """Base class for calling tools from rapidapi hub: https://rapidapi.com/hub
    """
    def __init__(self):
        super().__init__()
        self.url: str = None
        self.host_name: str = None
        self.api_key: str = None

    def run(self, params: dict):
        pass

    def get_tool_call_format(self) -> dict:
        pass


class BaseHuggingfaceTool(BaseTool):
    """Base class for calling models from huggingface

    """
    def __init__(self):
        super().__init__()
        self.url: str = None
        self.host_name: str = None
        self.api_key: str = None

    def run(self, params: dict):
        pass

    def get_tool_call_format(self) -> dict:
        pass



import argparse
import json
import subprocess
import requests
import gzip
import base64
import sys
import os


class Interactor:
    def __init__(self):
        script_path = os.path.abspath(__file__)
        script_dir = os.path.dirname(script_path)
        self.base_folder = script_dir

    def list_available_agents(self) -> list:
        """List available agents in the database"""
        url = "https://openagi-beta.vercel.app/api/get_all_agents"
        response = requests.get(url)
        response: dict = response.json()
        agent_list = []
        for v in list(response.values())[:-1]:
            agent_list.append({
                "agent": "/".join([v["author"], v["name"]])
            })
        return agent_list

    def download_agent(self, agent: str) -> None:
        """Download an agent from the database

        Args:
            agent (str): in the format of "author/agent_name"
        """
        assert "/" in agent, 'agent_name should in the format of "author/agent_name"'
        author, name = agent.split("/")
        # print(author, name)
        query = f'https://openagi-beta.vercel.app/api/download?author={author}&name={name}'
        response = requests.get(query)
        response: dict = response.json()

        agent_folder = os.path.join(self.base_folder, agent)

        if not os.path.exists(agent_folder):
            os.makedirs(agent_folder)

        encoded_config = response.get('config')
        encoded_code = response.get("code")
        encoded_reqs = response.get('dependencies')

        self.download_config(
            self.decompress(encoded_config),
            agent
        )
        self.download_code(
            self.decompress(encoded_code),
            agent
        )
        self.download_reqs(
            self.decompress(encoded_reqs),
            agent
        )

    def upload_agent(self, agent) -> None:
        """Upload an agent to the database

        Args:
            agent (str): in the format of "author/agent_name"
        """
        agent_dir = os.path.join(self.base_folder, agent)

        author, name = agent.split("/")

        config_file = os.path.join(agent_dir, "config.json")
        with open(config_file, 'r') as f:
            config_data: dict[str, dict] = json.load(f)

        meta_data = config_data.get('meta')

        headers = { "Content-Type": "application/json" }

        # compress python code
        encoded_code = self.compress(
            self.minify_python_code(agent_dir)
        )

        # compress meta_requirements.txt
        encoded_reqs =  self.compress(
            self.minify_reqs(agent_dir)
        )

        # compress config.json
        config_str = json.dumps(config_data)
        encoded_config = self.compress(config_str)

        upload_content = {
            'author': author,
            'name': name,
            'version': meta_data.get('version'),
            'license': meta_data.get('license'),
            'config': encoded_config,
            'code': encoded_code,
            'dependencies': encoded_reqs
        }

        url = 'https://openagi-beta.vercel.app/api/upload'

        response = requests.post(
            url,
            data=json.dumps(upload_content),
            headers=headers
        )
        if response.content:
            print("Uploaded successfully.")

    def minify_python_code(self, agent_dir):
        code_path = os.path.join(agent_dir, "agent.py")
        with open(code_path, 'r') as file:
            lines: list[str] = file.readlines()
        minified_lines = []
        for line in lines:
            stripped_line = line.rstrip()
            if stripped_line and not stripped_line.lstrip().startswith("#"):
                minified_lines.append(stripped_line)
        minified_code = "\n".join(minified_lines)
        return minified_code

    def minify_reqs(self, agent_dir):
        req_path = os.path.join(agent_dir, "meta_requirements.txt")
        with open(req_path, 'r') as file:
            self.reqs: str = file.read()
        cleaned = [line.strip() for line in self.reqs.split(
            '\n') if line.strip() and not line.startswith('#')]
        minified_reqs = ';'.join(cleaned)
        return minified_reqs

    def minify_config(self, config_data):
        minified_config = self.compress(config_data)
        return minified_config

    def compress(self, minified_data):
        compressed_data = gzip.compress(minified_data.encode('utf-8'))
        encoded_data = base64.b64encode(compressed_data)
        encoded_data = encoded_data.decode('utf-8')
        return encoded_data


    def decompress(self, encoded_data):
        compressed_data = base64.b64decode(encoded_data)
        decompressed_data = gzip.decompress(compressed_data)
        decompressed_data = decompressed_data.decode("utf-8")
        decompressed_data.replace(";", "\n")
        return decompressed_data

    def download_config(self, config_data, agent) :
        config_path = os.path.join(self.base_folder, agent, "config.json")
        config_data = json.loads(config_data)
        with open(config_path, "w") as w:
            json.dump(config_data, w, indent=4)

    def download_reqs(self, reqs_data, agent):
        reqs_path = os.path.join(self.base_folder, agent, "meta_requirements.txt")

        reqs_data = reqs_data.replace(";", "\n")

        with open(reqs_path, 'w') as file:
            file.write(reqs_data)

    def download_code(self, code_data, agent):
        code_path = os.path.join(self.base_folder, agent, "agent.py")

        with open(code_path, 'w', newline='') as file:
            file.write(code_data)

    def check_reqs_installed(self, agent):
    # Run the `conda list` command and capture the output
        reqs_path = os.path.join(self.base_folder, agent, "meta_requirements.txt")

        result = subprocess.run(['conda', 'list'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Decode the output from bytes to string
        with open(reqs_path, "r") as f:
            reqs = []
            lines = f.readlines()
            for line in lines:
                line = line.replace("\n", "")
                if "==" in line:
                    reqs.append(line.split("==")[0])
                else:
                    reqs.append(line)

        output = result.stdout.decode('utf-8')

        # Extract the list of installed packages
        installed_packages = [line.split()[0] for line in output.splitlines() if line]

        # Check for each package if it is installed
        for req in reqs:
            if req not in installed_packages:
                return False

        return True


    def install_agent_reqs(self, agent):
        reqs_path = os.path.join(self.base_folder, agent, "meta_requirements.txt")
        subprocess.check_call([
            sys.executable,
            "-m",
            "pip",
            "install",
            "-r",
            reqs_path
        ])

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", choices=["download", "upload"])
    parser.add_argument("--agent", required=True)
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    pass
    # list_available_agents() # list agents that can be used from db

    args = parse_args()
    mode = args.mode
    agent = args.agent

    client = Interactor()
    # client.check_reqs_installed(agent)
    # client = Interactor()
    if mode == "download":
        client.download_agent(agent) # download agents

    else:
        assert mode == "upload"
        client.upload_agent(agent) # upload agents



from ..base import BaseTool
# from langchain_core.documents import Document
from typing import Optional, Any, Dict
class Wikipedia(BaseTool):
    """Wikipedia tool, refactored from langchain.

    To use, you should have the ``wikipedia`` python package installed.
    This wrapper will use the Wikipedia API to conduct searches and
    fetch page summaries. By default, it will return the page summaries
    of the top-k results.
    It limits the Document content by doc_content_chars_max.
    """
    def __init__(self):
        super().__init__()
        self.WIKIPEDIA_MAX_QUERY_LENGTH = 300
        self.top_k_results = 3
        self.lang = "en"
        self.load_all_available_meta: bool = False
        self.doc_content_chars_max: int = 4000
        self.wiki_client = self.build_client()

    def build_client(self):
        try:
            import pyopenagi.tools.wikipedia.wikipedia as wikipedia

            wikipedia.set_lang(self.lang)

        except ImportError:
            raise ImportError(
                "Could not import wikipedia python package. "
                "Please install it with `pip install wikipedia`."
            )
        return wikipedia

    def run(self, query: Dict[str, str]) -> str:
        """Run Wikipedia search and get page summaries."""
        if not isinstance(query, dict) or 'query' not in query:
            raise TypeError("Query must be a dictionary with a 'query' key")
        query_str = query['query'][:self.WIKIPEDIA_MAX_QUERY_LENGTH]  # Extract and slice the query string
        page_titles = self.wiki_client.search(query_str, results=self.top_k_results)
        summaries = []
        for page_title in page_titles[: self.top_k_results]:
            if wiki_page := self._fetch_page(page_title):
                if summary := self._formatted_page_summary(page_title, wiki_page):
                    summaries.append(summary)
        if not summaries:
            return "No good Wikipedia Search Result was found"
        return "\n\n".join(summaries)[: self.doc_content_chars_max]

    @staticmethod
    def _formatted_page_summary(page_title: str, wiki_page: Any) -> Optional[str]:
        return f"Page: {page_title}\nSummary: {wiki_page.summary}"

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "wikipedia",
				"description": "Provides relevant information about the destination",
				"parameters": {
					"type": "object",
					"properties": {
						"query": {
							"type": "string",
							"description": "Search query for Wikipedia"
						}
					},
					"required": [
						"query"
					]
				}
			}
		}
        return tool_call_format



# make sure we can create agents

from pyopenagi.agents.agent_process import AgentProcess
from pyopenagi.utils.chat_template import Query

def test_agent_creation():
    agent_process = AgentProcess(
        agent_name="example/academic_agent",
        query=Query(
            messages = [
                {"role": "user", "content": "Summarize researches of quantum computing in recent five years."}
            ]
        )
    )
    # Use plain assert statements for testing conditions
    assert agent_process.agent_name == "example/academic_agent", "Agent name does not match"
    # Add more assertions here as necessary to validate the properties of the agent_process object



from diffusers import AutoPipelineForText2Image
import torch

from ..base import BaseHuggingfaceTool

class SdxlTurbo(BaseHuggingfaceTool):
    def __init__(self):
        super().__init__()
        self.pipe = AutoPipelineForText2Image.from_pretrained("stabilityai/sdxl-turbo", torch_dtype=torch.float16, variant="fp16")

    def run(self, params):
        prompt = params["prompt"]
        self.pipe.to("cuda")
        image = self.pipe(prompt=prompt, num_inference_steps=1, guidance_scale=0.0).images[0]
        return image

    def get_tool_call_format(self):
        tool_call_format = {
            "type": "function",
            "function": {
                "name": "sdxl_turbo",
                "description": "generate images with the given texts",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "prompt": {
                            "type": "string",
                            "description": "prompt description of the image to be generated"
                        }
                    },
                    "required": [
                        "prompt"
                    ]
                }
            }
        }
        return tool_call_format



import os
import pytest

from pyopenagi.tools.wolfram.wolfram_alpha import WolframAlpha
from dotenv import load_dotenv, find_dotenv

@pytest.fixture(scope="module")
def test_wolfram_alpha_id():
    load_dotenv(find_dotenv())
    if "WOLFRAM_ALPHA_APPID" not in os.environ or not os.environ["WOLFRAM_ALPHA_APPID"]:
        with pytest.raises(ValueError):
            WolframAlpha()
        pytest.skip("WolframAlpha app id is not set.")
    else:
        return True

@pytest.mark.usefixtures("test_wolfram_alpha_id")
def test_wolfram_alpha():
    wolfram_alpha = WolframAlpha()
    query = "What is the square root of 144?"
    result = wolfram_alpha.run(query)
    assert "12" in result



import os
import pytest
import requests
from requests.models import Response
import json

from pyopenagi.tools.imdb.top_series import TopSeries
from dotenv import load_dotenv, find_dotenv

@pytest.fixture(scope="module")
def test_rapid_api_key():
    load_dotenv(find_dotenv())
    if "RAPID_API_KEY" not in os.environ or not os.environ["RAPID_API_KEY"]:
        with pytest.raises(ValueError):
            TopSeries()
        pytest.skip("RAPID api key is not set.")
    else:
        return True

class ImdbTopSeriesMock:
    @staticmethod
    def json():
        mock_items = []
        mock_title = "Mock Title"
        mock_description = "Mock Description."
        mock_image = "https://m.media-amazon.com/images/Mock/Standard.Image.jpg"
        mock_big_image = "https://m.media-amazon.com/images/Mock/Big.Image.jpg"
        mock_genre = ["Drama", "Fantasy"]
        mock_thumbnail = "https://m.media-amazon.com/images/Mock/Thumb.Image.jpg"
        mock_rating = 9.2
        mock_year = "2011-2019"
        mock_imdbid = "tt0000000"
        mock_imdb_link = "https://www.imdb.com/title/tt0000000"

        for i in range(100):
            mock_items.append(
                {
                    "rank": i + 1,
                    "title": mock_title,
                    "description": mock_description,
                    "image": mock_image,
                    "big_image": mock_big_image,
                    "genre": mock_genre,
                    "thumbnail": mock_thumbnail,
                    "rating": mock_rating,
                    "id": f"top{i+1}",
                    "year": mock_year,
                    "imdbid": mock_imdbid,
                    "mock_imdb_link": mock_imdb_link,
                }
            )
        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = str.encode(json.dumps(mock_items))
        return mock_response.json()


@pytest.fixture(autouse=True)
def mock_response(monkeypatch):
    def mock_get(*args, **kwargs):
        return ImdbTopSeriesMock()

    monkeypatch.setattr(requests, "get", mock_get)


@pytest.mark.usefixtures("test_rapid_api_key")
@pytest.mark.parametrize(
    "valid_start, valid_end",
    [
        [1, 100],
        [60, 61],
        [60, 62],
    ],
)
def test_top_series_api_valid_input_outputs_valid_delimiter_count(
    valid_start, valid_end
):
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    params = {"start": valid_start, "end": valid_end}
    result = top_series_api.run(params=params)
    assert isinstance(result, str)
    assert result.count(";") == max(0, int(valid_end) - int(valid_start))

@pytest.mark.usefixtures("test_rapid_api_key")
def test_top_series_api_reverse_range_returns_blank():
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    params = {"start": 100, "end": 0}
    result = top_series_api.run(params=params)
    assert result == "Top 100-0 series ranked by IMDB are: "


@pytest.mark.parametrize(
    "invalid_start, valid_end",
    [
        ["0", 100],
        [0.5, 100],
        [[], 100],
        [{}, 100]
    ]
)
@pytest.mark.usefixtures("test_rapid_api_key")
def test_top_series_api_invalid_start_type_raises_typeerror(invalid_start, valid_end):
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    params = {"start": invalid_start, "end": valid_end}
    with pytest.raises(TypeError):
        top_series_api.run(params=params)


@pytest.mark.parametrize(
    "invalid_start, valid_end",
    [
        [1, "0"],
        [1, 0.5],
        [1, []],
        [1, {}]
    ]
)
@pytest.mark.usefixtures("test_rapid_api_key")
def test_top_series_api_invalid_end_type_raises_typeerror(invalid_start, valid_end):
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    params = {"start": invalid_start, "end": valid_end}
    with pytest.raises(TypeError):
        top_series_api.run(params=params)

@pytest.mark.usefixtures("test_rapid_api_key")
def test_top_series_api_invalid_start_count_raises_indexerror():
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    invalid_start = {"start": 101, "end": 102}
    with pytest.raises(IndexError):
        top_series_api.run(params=invalid_start)

@pytest.mark.usefixtures("test_rapid_api_key")
def test_top_series_api_invalid_end_count_raises_indexerror():
    load_dotenv(find_dotenv())
    top_series_api = TopSeries()
    invalid_end = {"start": 1, "end": 101}
    with pytest.raises(IndexError):
        top_series_api.run(params=invalid_end)



import os
import pytest

from pyopenagi.tools.currency_converter.currency_converter import CurrencyConverter
from dotenv import load_dotenv, find_dotenv

@pytest.fixture(scope="module")
def test_rapid_api_key():
    load_dotenv(find_dotenv())
    if "RAPID_API_KEY" not in os.environ or not os.environ["RAPID_API_KEY"]:
        with pytest.raises(ValueError):
            CurrencyConverter()
        pytest.skip("Rapid api key is not set.")
    else:
        return True

@pytest.mark.usefixtures("test_rapid_api_key")
def test_currency_converter_api():
    load_dotenv(find_dotenv())
    currency_converter_api = CurrencyConverter()
    params = {
        "from": "USD",
        "to": "EUR",
        "amount": 2
    }
    result = currency_converter_api.run(params=params)
    print(result)
    assert isinstance(result, str)



from ..base import BaseRapidAPITool

from typing import Any, Dict, List, Optional

# from pydantic import root_validator

from pyopenagi.utils.utils import get_from_env

import requests

class TopMovieAPI(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://imdb-top-100-movies.p.rapidapi.com/"
        self.host_name = "imdb-top-100-movies.p.rapidapi.com"

        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params):
        start = int(params["start"]) if "start" in params else 1
        end = int(params["end"])
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        response = requests.get(self.url, headers=headers).json()
        result = self.parse_result(response, start, end)
        return result


    def parse_result(self, response, start, end) -> str:
        result = []
        # print(response)
        for i in range(start, end):
            item = response[i]
            result.append(f'{item["title"]}, {item["genre"]}, {item["rating"]}, published in {item["year"]}')

        return f"Top {start}-{end} series ranked by IMDB are: " + ";".join(result)

    def get_tool_call_format(self):
        tool_call_format = {
            "type": "function",
            "function": {
                "name": "top_movies",
                "description": "Query the latest top start-to-end movies ranked by Imdb",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "start": {
                            "type": "string",
                            "description": "start of the rank range of the Imdb movies",
                            "default": "1"
                        },
                        "end": {
                            "type": "string",
                            "description": "end of the rank range of the Imdb movies"
                        }
                    },
                    "required": [
                        "end"
                    ]
                }
            }
        }
        return tool_call_format



# contains tools for logging actions
# the thread scheduler, agents, and the LLM kernel of AIOS are all loggable here

import click

import os

from datetime import datetime

class BaseLogger:
    def __init__(self,
            logger_name,
            log_mode = "console",
        ) -> None:
        self.logger_name = logger_name
        self.log_mode = log_mode
        self.log_file = self.load_log_file() if log_mode == "file" else None

        self.level_color = dict()

    def log(self, content, level):
        if self.log_mode == "console":
            self.log_to_console(content, level)
        else:
            assert self.log_mode == "file" and self.log_file is not None
            self.log_to_file(content, self.log_file)

    def load_log_file(self):
        pass

    def log_to_console(self, content, level):
        # print(content)
        click.secho(f"[{self.logger_name}] " + content, fg=self.level_color[level])

    def log_to_file(self, content, log_file):
        with open(log_file, "a") as w:
            w.writelines(content)

class SchedulerLogger(BaseLogger):
    """ log the threads """

    def __init__(self, logger_name, log_mode="console") -> None:
        super().__init__(logger_name, log_mode)
        self.level_color = {
            "execute": "green",
            "suspend": "yellow",
            "info": "white"
        }

    def load_log_file(self):
        date_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log_dir = os.path.join(os.getcwd(), "logs", "scheduler")
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_file = os.path.join(log_dir, f"{date_time}.txt")
        return log_file


class AgentLogger(BaseLogger):
    """ log the agents """

    def __init__(self, logger_name, log_mode="console") -> None:
        super().__init__(logger_name, log_mode)
        self.level_color = {
            "info": (248, 246, 227), # white
            "executing": (217, 237, 191), # green
            "suspending": (255, 235, 178), # yellow
            "done": (122, 162, 227) # blue
        }

    def load_log_file(self):
        date_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log_dir = os.path.join(os.getcwd(), "logs", "agents", self.logger_name)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_file = os.path.join(log_dir, f"{date_time}.txt")
        return log_file


class LLMKernelLogger(BaseLogger):
    """ log the LLM kernel """

    def __init__(self, logger_name, log_mode="console") -> None:
        super().__init__(logger_name, log_mode)
        self.level_color = {
            "info": (246, 245, 242),
            "executing": (65, 176, 110), # green
            "suspending": (255, 201, 74), # yellow
            "done": (122, 162, 227) # blue
        }

    def log_to_console(self, content, level):
        # print(content)
        click.secho(f"[\U0001F916{self.logger_name}] " + content, fg=self.level_color[level], bold=True)

    def load_log_file(self):
        date_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log_dir = os.path.join(os.getcwd(), "logs", "llm_kernel", self.logger_name)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_file = os.path.join(log_dir, f"{date_time}.txt")
        return log_file



import requests

from ..base import BaseTool

from typing import List
# from pydantic import root_validator

from pyopenagi.utils.utils import get_from_env

class BingSearch(BaseTool):
    """Bing Search Tool, refactored from langchain.
    In order to set this up, follow instructions at:
    https://levelup.gitconnected.com/api-tutorial-how-to-use-bing-web-search-api-in-python-4165d5592a7e
    """
    def __init__(self):
        super().__init__()
        self.url = "https://api.bing.microsoft.com/v7.0/search" # temporarily
        self.bing_subscription_key = get_from_env("BING_SUBSCRIPTION_KEY")
        self.k: int = 10 # topk searched results
        # search_kwargs: dict

    def _bing_search_results(self, search_term: str, count: int) -> List[dict]:
        headers = {"Ocp-Apim-Subscription-Key": self.bing_subscription_key}
        params = {
            "q": search_term,
            "count": count,
            "textDecorations": True,
            "textFormat": "HTML",
            # **self.search_kwargs,
        }
        response = requests.get(
            self.bing_search_url,
            headers=headers,
            params=params,  # type: ignore
        )
        response.raise_for_status()
        search_results = response.json()
        if "webPages" in search_results:
            return search_results["webPages"]["value"]
        return []

    def run(self, query: str) -> str:
        """Run query through BingSearch and parse result."""
        response = self._bing_search_results(query, count=self.k)
        result = self.parse_result(response)
        return result

    def parse_result(self, response):
        snippets = []
        if len(response) == 0:
            return "No good Bing Search Result was found"
        for result in response:
            snippets.append(result["snippet"])

        return " ".join(snippets)



from ..base import BaseRapidAPITool

# from pydantic import root_validator

from pyopenagi.utils.utils import get_from_env

import requests

SUPPORTED_APIS = [
    "typeOf", "hasTypes", "partOf", "hasParts",
    "instanceOf", "hasInstances", "similarTo",
    "also", "entails", "memberOf", "hasMembers",
    "substanceOf", "hasSubstances", "inCategory",
    "hasCategories", "usageOf", "hasUsages",
    "inRegion", "regionOf", "pertainsTo", "synonyms",
    "examples", "antonyms", "pronunciation",
]

class WordsAPI(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.base_url = "https://wordsapiv1.p.rapidapi.com/words/"
        self.url = self.base_url
        self.host_name = "wordsapiv1.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        try:
            self.word = params["word"]
            self.api_name = params["api_name"]
        except KeyError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for words api. "
                "Please make sure it contains two keys: 'words' and 'api_name'"
            )

        if not self.is_supported(self.api_name):
            raise ValueError(
                f"{self.api_name} is currently not supported!"
            )

        self.url = f"{self.base_url}{self.word}/{self.api_name}"
        response = requests.get(self.url, headers=headers).json()
        result = self.parse_result(response)
        return result

    def parse_result(self, response) -> str:
        # fail response: {'success': False, 'message': 'word not found'}
        if "success" in response and not response["success"]:
            return response["message"]

        return response["word"] + " " + self.api_name + " [" + ",".join(response[self.api_name]) + "]"

    def is_supported(self, api_name):
        return api_name in SUPPORTED_APIS



from ...react_agent import ReactAgent

class AcademicAgent(ReactAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):
        ReactAgent.__init__(self, agent_name, task_input, agent_process_factory, log_mode)
        self.workflow_mode = "automatic"

    def manual_workflow(self):
        pass

    def run(self):
        return super().run()



import os
import pytest

from pyopenagi.tools.words_api.words_api import WordsAPI
from dotenv import load_dotenv, find_dotenv

@pytest.fixture(scope="module")
def test_rapid_api_key():
    load_dotenv(find_dotenv())
    if "RAPID_API_KEY" not in os.environ or not os.environ["RAPID_API_KEY"]:
        with pytest.raises(ValueError):
            WordsAPI()
        pytest.skip("Rapid api key is not set.")

@pytest.mark.usefixtures("test_rapid_api_key")
def test_words_api():
    words_api = WordsAPI()
    params = {
        "word": "look",
        "api_name": "typeOf",
    }
    result = words_api.run(params=params)
    print(result)
    assert isinstance(result, str)



from ..base import BaseTool

from pyopenagi.utils.utils import get_from_env
class WolframAlpha(BaseTool):
    """Wolfram Alpha Tool, refactored from langchain.

    Docs for using:

    1. Go to wolfram alpha and sign up for a developer account
    2. Create an app and get your APP ID
    3. Save your APP ID into WOLFRAM_ALPHA_APPID env variable
    4. pip install wolframalpha

    """
    def __init__(self):
        super().__init__()
        self.wolfram_alpha_appid = get_from_env("WOLFRAM_ALPHA_APPID")
        self.wolfram_client = self.build_client()

    def build_client(self):
        try:
            import wolframalpha

        except ImportError:
            raise ImportError(
                "wolframalpha is not installed. "
                "Please install it with `pip install wolframalpha`"
            )
        client = wolframalpha.Client(self.wolfram_alpha_appid)
        return client

    def run(self, query: str) -> str:
        """Run query through WolframAlpha and parse result."""
        res = self.wolfram_client.query(query)

        try:
            assumption = next(res.pods).text
            answer = next(res.results).text
        except StopIteration:
            return "Wolfram Alpha wasn't able to answer it"

        if answer is None or answer == "":
            # We don't want to return the assumption alone if answer is empty
            return "No good Wolfram Alpha Result was found"
        else:
            return f"Assumption: {assumption} \nAnswer: {answer}"


    def get_tool_call_format(self):
        tool_call_format = {
            "type": "function",
            "function": {
                "name": "wolfram_alpha",
                "description": "Use specific mathematical knowledge (algebra, calculus, geometry, etc) to answer the given query",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "the abstracted mathematical query that needs to be answered"
                        }
                    },
                    "required": [
                        "query"
                    ]
                }
            }
        }
        return tool_call_format



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

import json

class RestaurantLocationSearch(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://tripadvisor16.p.rapidapi.com/api/v1/restaurant/searchLocation"
        self.host_name = "tripadvisor16.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        try:
            self.query_string = {
                "query": params["query"],
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for tripadvisor search restaurant location api. "
                "Please make sure it contains the key: 'query'"
            )

        # print(self.query_string)

        response = requests.get(self.url, headers=headers, params=self.query_string).json()
        return self.parse_result(response)

    def parse_result(self, response) -> str:
        limited_results = response['data'][:2]

        simplified_results = []
        for result in limited_results:
            simplified_result = {
                'locationId': result['locationId'],
                'localizedName': result['localizedName'],
                'latitude': result['latitude'],
                'longitude': result['longitude']
            }
            simplified_results.append(simplified_result)

        return json.dumps(simplified_results)

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "restaurant_location_search",
				"description": "Search for a restaurant location by query",
				"parameters": {
					"type": "object",
					"properties": {
						"query": {
							"type": "string",
							"description": "Search query for restaurant location"
						}
					},
					"required": [
						"query"
					]
				}
			}
		}
        return tool_call_format



from ...react_agent import ReactAgent
class TravelAgent(ReactAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):
        ReactAgent.__init__(self, agent_name, task_input, agent_process_factory, log_mode)

    def automatic_workflow(self):
        return super().automatic_workflow()

    def manual_workflow(self):
        workflow = [
            {
                "message": "identify the destination and search for hotel locations",
                "tool_use": ["hotel_location_search"]
            },
            {
                "message": "based on the hotel locations, find suitable hotels using the hotel_search tool, and select the best one. ",
                "tool_use": None
            },
            {
                "message": "get detailed information about the selected hotel",
                "tool_use": ["get_hotel_details"]
            },
            {
                "message": ["search for the nearest airport to the origin"],
                "tool_use": ["airport_search"]
            },
            {
                "message": ["search for the nearest airport to the destination"],
                "tool_use": ["airport_search"]
            },
            {
                "message": ["find available flights to the destination airport using the correct date"],
                "tool_use": ["flight_search"]
            },
            {
                "message": ["search for restaurant locations near destination"],
                "tool_use": ["restaurant_location_search"]
            },
            {
                "message": ["based on the restaurant locations, find suitable restaurants"],
                "tool_use": ["restaurant_search"]
            },
            {
                "message": ["get detailed information about the selected restaurants"],
                "tool_use": ["get_restaurant_details"]
            },
            {
                "message": ["Gather additional relevant information about the destination the user is visiting"],
                "tool_use": ["wikipedia"]
            },
            {
                "message": ["integrate the information gathered from the previous steps to provide a comprehensive travel plan"],
                "tool_use": None
            }
        ]
        return workflow

    def run(self):
        return super().run()



class Query:
    def __init__(self,
            messages,
            tools = None,
            message_return_type = "text"
        ) -> None:
        """Query format

        Args:
            messages (list):
            [
                {"role": "xxx", content_key: content_value}
            ]
            tools (optional): tools that are used for function calling. Defaults to None.
        """
        self.messages = messages
        self.tools = tools
        self.message_return_type = message_return_type

class Response:
    def __init__(
            self,
            response_message,
            tool_calls: list = None
        ) -> None:
        """Response format

        Args:
            response_message (str): "generated_text"
            tool_calls (list, optional):
            [
                {"name": "xxx", "parameters": {}}
            ].
            Default to None.
        """
        self.response_message = response_message
        self.tool_calls = tool_calls



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

class CurrencyConverter(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://currency-converter5.p.rapidapi.com/currency/convert"
        self.host_name = "currency-converter5.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        try:
            self.query_string = {
                "from": params["from"],
                "to": params["to"],
                "amount": params["amount"] if "amount" in params else "1.0"
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for currency converter api. "
                "Please make sure it contains two keys: 'from' and 'to'"
            )

        response = requests.get(self.url, headers=headers, params=self.query_string).json()

        result = self.parse_result(response)
        return result

    def parse_result(self, response) -> str:
        results = []
        amount = str(response["amount"])
        base = response["base_currency_name"]
        rates = response["rates"]

        for key, value in rates.items():
            converted = value["currency_name"]
            converted_rate = value["rate"]
            converted_amount = value["rate_for_amount"]
            results.append("Currency from " + base + " to " + converted + "is " + converted_rate + ".", )
            results.append(amount + " " + base + "can be converted to " + converted_amount + " " + converted + ".")

        return " ".join(results)

    def get_tool_call_format(self):
        tool_call_format = {
            "type": "function",
            "function": {
                "name": "currency_converter",
                "description": "Provides currency exchange rates convert base currency to desired currency with the given amount",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "from": {
                            "type": "string",
                            "description": "Base currency code, e.g., AUD, CAD, EUR, GBP..."
                        },
                        "to": {
                            "type": "string",
                            "description": "Desired currency code, e.g., AUD, CAD, EUR, GBP..."
                        },
                        "amount": {
                            "type": "string",
                            "default": "1.0",
                            "description": "The amount to be converted"
                        }
                    },
                    "required": [
                        "from",
                        "to"
                    ]
                }
            }
        }
        return tool_call_format



from ...react_agent import ReactAgent
class CreationAgent(ReactAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):
        ReactAgent.__init__(self, agent_name, task_input, agent_process_factory, log_mode)
        self.workflow_mode = "automatic"

    def automatic_workflow(self):
        return super().automatic_workflow()

    def manual_workflow(self):
        workflow = [
            {
                "message": "filled in more details about the user's requirement",
                "tool_use": None
            },
            {
                "message": "generate an image based on the user's requirements and filled details",
                "tool_use": ["sdxl-turbo"]
            }
        ]
        return workflow

    def run(self):
        return super().run()



# make a queue for agents to use

import queue

class BaseQueue:

    _queue = queue.Queue()

    @classmethod
    def add_message(cls, message):
        cls._queue.put(message)

    @classmethod
    def get_message(cls):
        return cls._queue.get(block=True, timeout=1)

    @classmethod
    def is_empty(cls):
        return cls._queue.empty()



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

import json

class HotelSearch(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://tripadvisor16.p.rapidapi.com/api/v1/hotels/searchHotels"
        self.host_name = "tripadvisor16.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }

        try:
            self.query_string = {
                "geoId": params["geoId"],
                "checkIn": params["checkIn"],
                "checkOut": params["checkOut"],
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for tripadvisor search hotel api. "
                "Please make sure it contains following required keys: "
                "geoId",
                "checkIn",
                "checkOut",
            )

        response = requests.get(self.url, headers=headers, params=self.query_string).json()
        return self.parse_result(response)


    def parse_result(self, response) -> str:
        if 'data' in response and 'data' in response['data']:
            hotels_data = response['data']['data'][:2]
            relevant_info = []
            for hotel in hotels_data:
                relevant_info.append({
                    'id': hotel['id'],
                    'title': hotel['title'],
                    'secondaryInfo': hotel['secondaryInfo'],
                    'bubbleRating': hotel['bubbleRating'],
                    'priceForDisplay': hotel['priceForDisplay'],
                    'priceDetails': hotel['priceDetails'],
                    'priceSummary': hotel['priceSummary']
                })
            return json.dumps(relevant_info)
        else:
            return json.dumps([])

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "hotel_search",
				"description": "Provides details about a hotel",
				"parameters": {
					"type": "object",
					"properties": {
						"geoId": {
							"type": "string",
							"description": "The geoId of the hotel to search for"
						},
						"checkIn": {
							"type": "string",
							"format": "date",
							"description": "The check in date"
						},
						"checkOut": {
							"type": "string",
							"format": "date",
							"description": "The check out date"
						}
					},
					"required": [
						"geoId",
						"checkIn",
						"checkOut"
					]
				}
			}
		}
        return tool_call_format



# helper utils AIOS uses
# file is not particularly organized but mainly small functions

import argparse

import os
import shutil

import json

from typing import Dict, Any, Optional

import re

# logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
# logger = logging.getLogger(__name__)

def parse_global_args():
    parser = argparse.ArgumentParser(description="Parse global parameters")
    parser.add_argument('--llm_name', type=str, default="gemma-2b-it", help="Specify the LLM name of AIOS")
    parser.add_argument('--max_gpu_memory', type=json.loads, help="Max gpu memory allocated for the LLM")
    parser.add_argument('--eval_device', type=str, help="Evaluation device")
    parser.add_argument('--max_new_tokens', type=int, default=256, help="The maximum number of new tokens for generation")
    parser.add_argument("--scheduler_log_mode", type=str, default="console", choices=["console", "file"])
    parser.add_argument("--agent_log_mode", type=str, default="console", choices=["console", "file"])
    parser.add_argument("--llm_kernel_log_mode", type=str, default="console", choices=["console", "file"])

    return parser

def extract_before_parenthesis(s: str) -> str:
    match = re.search(r'^(.*?)\([^)]*\)', s)
    return match.group(1) if match else s

def get_from_dict_or_env(
    data: Dict[str, Any], key: str, env_key: str, default: Optional[str] = None
) -> str:
    """Get a value from a dictionary or an environment variable."""
    if key in data and data[key]:
        return data[key]
    else:
        return get_from_env(key, env_key, default=default)


def get_from_env(env_key: str, default: Optional[str] = None) -> str:
    """Get a value from an environment variable."""
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    elif default is not None:
        return default
    else:
        raise ValueError(
            f"Did not find {env_key}, please add an environment variable"
            f" `{env_key}` which contains it. "
        )

class Logger:
    def __init__(self, log_mode) -> None:
        self.log_mode = log_mode

    def log(self, info, path=None):
        if self.log_mode == "console":
            print(info)
        else:
            assert self.log_mode == "file"
            with open(path, "w") as w:
                w.write(info + "\n")

def delete_directories(root_dir, target_dirs):
    """
    Recursively deletes directories with names in target_dirs starting from root_dir.
    """
    for dirpath, dirnames, filenames in os.walk(root_dir, topdown=False):
        for dirname in dirnames:
            if dirname in target_dirs:
                full_path = os.path.join(dirpath, dirname)
                # print(f"Deleting {full_path}...")
                shutil.rmtree(full_path, ignore_errors=True)



from ...base_agent import BaseAgent

import time

import argparse

from ....utils import Message

from pathlib import PurePosixPath
import os
import chromadb
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.core import VectorStoreIndex, SimpleDirectoryReader
from llama_index.core import StorageContext
from llama_index.core import PromptTemplate
from llama_index.core.retrievers import VectorIndexRetriever
from llama_index.vector_stores.chroma import ChromaVectorStore

from openagi.src.agents.base import BaseAgent

class RAGAgent(BaseAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 llm,
                 agent_process_queue,
                 agent_process_factory,
                 log_mode: str
        ):
        BaseAgent.__init__(self, agent_name, task_input, llm, agent_process_queue, agent_process_factory, log_mode)

    def run(self):
        request_waiting_times = []
        request_turnaround_times = []
        query = self.task_input

        self.logger.log(f"{query}\n", level="info")

        context = self.retrive(query)
        prompt = self.build_prompt(context_str=context, query_str=query)

        rounds = 0

        response, start_times, end_times, waiting_times, turnaround_times = self.get_response(
            message = Message(
                prompt = prompt,
                tools = None
            )
        )

        self.set_start_time(start_times[0])
        rounds += 1

        request_waiting_times.extend(waiting_times)
        request_turnaround_times.extend(turnaround_times)

        response_message = response.response_message

        self.logger.log(f"Final result is: {response.response_message}\n", level="info")
        final_result = response_message

        self.set_status("done")
        self.set_end_time(time=time.time())

        return {
            "agent_name": self.agent_name,
            "result": final_result,
            "rounds": rounds,
            "agent_waiting_time": self.start_time - self.created_time,
            "agent_turnaround_time": self.end_time - self.created_time,
            "request_waiting_times": request_waiting_times,
            "request_turnaround_times": request_turnaround_times,
        }

    def retrive(self, query: str):
        script_dir = os.path.dirname(os.path.realpath(__file__))
        self.data_path = PurePosixPath(script_dir, "data", "paul_graham").as_posix()
        self.db_path = PurePosixPath(script_dir, "chroma_db").as_posix()
        self.collection_name = "quickstart"
        self.embed_model = HuggingFaceEmbedding(model_name="BAAI/bge-base-en-v1.5")

        self.create_db_if_not_exists()

        db = chromadb.PersistentClient(path=self.db_path)
        chroma_collection = db.get_or_create_collection(self.collection_name)
        vector_store = ChromaVectorStore(chroma_collection=chroma_collection)
        index = VectorStoreIndex.from_vector_store(
            vector_store,
            embed_model=self.embed_model,
        )

        retriever = VectorIndexRetriever(index=index, similarity_top_k=2)
        retrieved_result = retriever.retrieve(query)
        context_str = retrieved_result[0].get_content()
        return context_str

    def create_db_if_not_exists(self):
        if os.path.exists(self.db_path):
            pass
        else:
            print("store documents to vector db!")
            documents = SimpleDirectoryReader(self.data_path).load_data()

            chroma_client = chromadb.PersistentClient(path=self.db_path)
            chroma_collection = chroma_client.create_collection(self.collection_name)
            vector_store = ChromaVectorStore(chroma_collection=chroma_collection)
            storage_context = StorageContext.from_defaults(vector_store=vector_store)
            index = VectorStoreIndex.from_documents(
                documents, storage_context=storage_context, embed_model=self.embed_model
            )
            index.storage_context.persist(persist_dir=self.db_path)

    def build_prompt(self, context_str: str, query_str: str):
        prompt_template_literal = (
            "{query_str}"

            "Use the following context as your learned knowledge, inside <context></context> XML tags.\n"
            "<context>\n"
                "{context_str}"
            "</context>\n"

            "Avoid mentioning that you obtained the information from the context.\n"
        )
        prompt_template = PromptTemplate(prompt_template_literal)
        final_prompt = prompt_template.format(context_str=context_str, query_str=query_str)
        return final_prompt

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Run RagAgent')
    parser.add_argument("--agent_name")
    parser.add_argument("--task_input")

    args = parser.parse_args()
    agent = RAGAgent(args.agent_name, args.task_input)
    agent.run()



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

import json

class FlightSearch(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://tripadvisor16.p.rapidapi.com/api/v1/flights/searchFlights"
        self.host_name = "tripadvisor16.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }

        try:
            self.query_string = {
                "sourceAirportCode": params["sourceAirportCode"],
                "date": params["date"],
                "destinationAirportCode": params["destinationAirportCode"],
                "itineraryType": params["itineraryType"],
                "sortOrder": params["sortOrder"],
                "classOfService": params["classOfService"],
                "returnDate": params["returnDate"]
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for tripadvisor search flight api. "
                "Please make sure it contains following required keys: "
                "sourceAirportCode",
                "destinationAirportCode",
                "itineraryType",
                "sortOrder",
                "classOfService",
                "returnDate",
                "date"
            )

        response = requests.get(self.url, headers=headers, params=self.query_string).json()
        return self.parse_result(response)

    def parse_result(self, response) -> str:
        # Accessing the 'flights' data from within the 'data' key
        if 'data' in response and 'flights' in response['data']:
            flights_data = response['data']['flights']
            simplified_results = []
            flight_count = 0
            for flight in flights_data:
                if flight_count >= 2:
                    break
                for segment in flight['segments']:
                    for leg in segment['legs']:
                        simplified_result = {
                            'originStationCode': leg['originStationCode'],
                            'destinationStationCode': leg['destinationStationCode'],
                            'departureDateTime': leg['departureDateTime'],
                            'arrivalDateTime': leg['arrivalDateTime'],
                            'classOfService': leg['classOfService'],
                            'marketingCarrierCode': leg['marketingCarrierCode'],
                            'operatingCarrierCode': leg['operatingCarrierCode'],
                            'flightNumber': leg['flightNumber'],
                            'numStops': leg['numStops'],
                            'distanceInKM': leg['distanceInKM'],
                            'isInternational': leg['isInternational']
                        }
                        simplified_results.append(simplified_result)
                flight_count += 1
            return json.dumps(simplified_results)
        else:
            return json.dumps([])

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "flight_search",
				"description": "Provides details about a flight",
				"parameters": {
					"type": "object",
					"properties": {
						"sourceAirportCode": {
							"type": "string",
							"description": "The source airport code of the flight to search for"
						},
						"date": {
							"type": "string",
							"format": "date",
							"description": "The date of the flight"
						},
						"returnDate": {
							"type": "string",
							"format": "date",
							"description": "The return date of the flight"
						},
						"destinationAirportCode": {
							"type": "string",
							"description": "The destination airport code of the flight"
						},
						"itineraryType": {
							"type": "string",
							"enum": [
								"ONE_WAY",
								"ROUND_TRIP"
							],
							"description": "The type of itinerary"
						},
						"sortOrder": {
							"type": "string",
							"enum": [
								"ML_BEST_VALUE",
								"DURATION",
								"PRICE",
								"EARLIEST_OUTBOUND_DEPARTURE",
								"EARLIEST_OUTBOUND_ARRIVAL",
								"LATEST_OUTBOUND_DEPARTURE",
								"LATEST_OUTBOUND_ARRIVAL"
							],
							"description": "The order to sort the results"
						},
						"classOfService": {
							"type": "string",
							"enum": [
								"ECONOMY",
								"PREMIUM_ECONOMY",
								"BUSINESS",
								"FIRST"
							],
							"description": "The class of service for the flight"
						},
						"numSeniors": {
							"type": "number",
							"description": "The number of seniors in the itinerary"
						},
						"numAdults": {
							"type": "number",
							"description": "The number of adults in the itinerary"
						}
					},
					"required": [
						"sourceAirportCode",
						"date",
						"destinationAirportCode",
						"itineraryType",
						"sortOrder",
						"classOfService",
						"numSeniors",
						"numAdults",
						"returnDate"
					]
				}
			}
		}
        return tool_call_format



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

class MoonPhaseSearch(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://moon-phase.p.rapidapi.com/basic"
        self.host_name = "moon-phase.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        response = requests.get(self.url, headers=headers)

        result = self.parse_result(response)
        return result

    def parse_result(self, response) -> str:
        return f'Current moon phase is {response["phase_name"]}. It has {response["days_until_next_full_moon"]} until next full moon. It has {response["days_until_next_new_moon"]} until the next new moon.'



from ...react_agent import ReactAgent

class MathAgent(ReactAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):
        ReactAgent.__init__(self, agent_name, task_input, agent_process_factory, log_mode)

    def automatic_workflow(self):
        return super().automatic_workflow()

    def manual_workflow(self):
        workflow = [
            {
                "message": "identify the tool to call to do some pre-calculation. ",
                "tool_use": ["wolfram_alpha", "currency_converter"]
            },
            {
                "message":"perform mathematical operations using the pre-calculated result, which could involve addition, subtraction, multiplication, or division with other numeric values to solve the problem.",
                "tool_use": None
            }
        ]
        return workflow

    def run(self):
        return super().run()



# uses zlibrary to compress data 
# same as util in AIOS

import zlib

class Compressor:
    def __init__(self) -> None:
        pass

    def compress(self, data):
        pass

    def decompress(self, compressed_data):
        pass

class ZLIBCompressor(Compressor):
    def __init__(self) -> None:
        pass

    def compress(self, data):
        compressed_data = zlib.compress(data.encode('utf-8'))
        return compressed_data

    def decompress(self, compressed_data):
        decompressed_data = zlib.decompress(compressed_data)
        return decompressed_data.decode('utf-8')



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

import json

class GetHotelDetails(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://tripadvisor16.p.rapidapi.com/api/v1/hotels/getHotelDetails"
        self.host_name = "tripadvisor16.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }

        try:
            self.query_string = {
                "id": params["id"],
                "checkIn": params["checkIn"],
                "checkOut": params["checkOut"],
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for tripadvisor get hotel details api. "
                "Please make sure it contains following required keys: "
                "id",
                "checkIn",
                "checkOut",
            )

        response = requests.get(self.url, headers=headers, params=self.query_string).json()
        return self.parse_result(response)

    def parse_result(self, response) -> str:
        if 'data' in response:
            hotel_data = response['data']
            relevant_info = {
                'name': hotel_data.get('title', ''),
                'rating': hotel_data.get('rating', ''),
                'address': hotel_data.get('location', {}).get('address', ''),
                'amenities': [amenity['title'] for amenity in hotel_data.get('about', {}).get('content', []) if amenity['title'] == 'Amenities'],
                'description': hotel_data.get('about', {}).get('content', [{}])[0].get('content', ''),
                'restaurantsNearby': [{
                    'title': hotel_data.get('restaurantsNearby', {}).get('content', [{}])[0].get('title', ''),
                    'rating': hotel_data.get('restaurantsNearby', {}).get('content', [{}])[0].get('bubbleRating', {}).get('rating', ''),
                    'primaryInfo': hotel_data.get('restaurantsNearby', {}).get('content', [{}])[0].get('primaryInfo', ''),
                    'distance': hotel_data.get('restaurantsNearby', {}).get('content', [{}])[0].get('distance', ''),
                }],
                'attractionsNearby': [{
                    'title': hotel_data.get('attractionsNearby', {}).get('content', [{}])[0].get('title', ''),
                    'rating': hotel_data.get('attractionsNearby', {}).get('content', [{}])[0].get('bubbleRating', {}).get('rating', ''),
                    'primaryInfo': hotel_data.get('attractionsNearby', {}).get('content', [{}])[0].get('primaryInfo', ''),
                    'distance': hotel_data.get('attractionsNearby', {}).get('content', [{}])[0].get('distance', ''),
                }]
            }
            return json.dumps(relevant_info)
        else:
            return json.dumps({})

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "get_hotel_details",
				"description": "Provides details about a hotel",
				"parameters": {
					"type": "object",
					"properties": {
						"id": {
							"type": "string",
							"description": "The ID of the hotel to get details for"
						},
						"checkIn": {
							"type": "string",
							"format": "date",
							"description": "The check in date"
						},
						"checkOut": {
							"type": "string",
							"format": "date",
							"description": "The check out date"
						}
					},
					"required": [
						"id",
						"checkIn",
						"checkOut"
					]
				}
			}
		}
        return tool_call_format



# stub implementation for agents
# defines a class standard each agent has to subclass

import os

import json

from .agent_process import (
    AgentProcess
)

import time

from threading import Thread

from ..utils.logger import AgentLogger

from ..utils.chat_template import Query

import importlib

from ..queues.llm_request_queue import LLMRequestQueue

class CustomizedThread(Thread):
    """ provides a specific agent runtime """

    def __init__(self, target, args=()):
        super().__init__()
        self.target = target
        self.args = args
        self.result = None

    def run(self):
        """
        instead of creating a Thread, it'll run a function and hold the value
        in BaseAgent, it's only usage, it creates a Thread and sets certain 
        values for the AgentProcess to provide runtime diagnostics
        """
        self.result = self.target(*self.args)

    def join(self):
        """ returns the result from the custom Thread runtime """
        super().join()
        return self.result

class BaseAgent:
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):

        self.agent_name = agent_name
        self.config = self.load_config()
        self.tool_names = self.config["tools"]

        self.agent_process_factory = agent_process_factory

        self.tool_list = dict()
        self.tools = []
        self.tool_info = [] # simplified information of the tool: {"name": "xxx", "description": "xxx"}

        self.load_tools(self.tool_names)

        self.start_time = None
        self.end_time = None
        self.request_waiting_times: list = []
        self.request_turnaround_times: list = []
        self.task_input = task_input
        self.messages = []
        self.workflow_mode = "manual" # (mannual, automatic)
        self.rounds = 0

        self.log_mode = log_mode
        self.logger = self.setup_logger()
        self.logger.log("Initialized. \n", level="info")

        self.set_status("active")
        self.set_created_time(time.time())


    def run(self):
        '''Execute each step to finish the task.'''
        pass

    # can be customization
    def build_system_instruction(self):
        pass

    def check_workflow(self, message):
        try:
            # print(f"Workflow message: {message}")
            workflow = json.loads(message)
            if not isinstance(workflow, list):
                return None

            for step in workflow:
                if "message" not in step or "tool_use" not in step:
                    return None

            return workflow

        except json.JSONDecodeError:
            return None

    def automatic_workflow(self):
        for i in range(self.plan_max_fail_times):
            response, start_times, end_times, waiting_times, turnaround_times = self.get_response(
                query = Query(
                    messages = self.messages,
                    tools = None,
                    message_return_type="json"
                )
            )

            if self.rounds == 0:
                self.set_start_time(start_times[0])

            self.request_waiting_times.extend(waiting_times)
            self.request_turnaround_times.extend(turnaround_times)

            workflow = self.check_workflow(response.response_message)

            self.rounds += 1

            if workflow:
                return workflow

            else:
                self.messages.append(
                    {
                        "role": "assistant",
                        "content": f"Fail {i+1} times to generate a valid plan. I need to regenerate a plan"
                    }
                )
        return None

    def manual_workflow(self):
        pass

    def snake_to_camel(self, snake_str):
        components = snake_str.split('_')
        return ''.join(x.title() for x in components)

    def load_tools(self, tool_names):

        for tool_name in tool_names:
            org, name = tool_name.split("/")
            module_name = ".".join(["pyopenagi", "tools", org, name])
            class_name = self.snake_to_camel(name)
            tool_module = importlib.import_module(module_name)
            tool_class = getattr(tool_module, class_name)
            self.tool_list[name] = tool_class()
            tool_format = tool_class().get_tool_call_format()
            self.tools.append(tool_format)
            self.tool_info.append(
                {"name": tool_format["function"]["name"], "description": tool_format["function"]["description"]}
            )

    def pre_select_tools(self, tool_names):
        pre_selected_tools = []
        for tool_name in tool_names:
            for tool in self.tools:
                if tool["function"]["name"] == tool_name:
                    pre_selected_tools.append(tool)
                    break

        return pre_selected_tools

    def setup_logger(self):
        logger = AgentLogger(self.agent_name, self.log_mode)
        return logger

    def load_config(self):
        """  
        loads each agent config with the values in the json
        such as the system prompt, functions 
        """
        script_path = os.path.abspath(__file__)
        script_dir = os.path.dirname(script_path)
        config_file = os.path.join(script_dir, self.agent_name, "config.json")
        with open(config_file, "r") as f:
            config = json.load(f)
            return config

    # the default method used for getting response from AIOS
    def get_response(self,
            query,
            temperature=0.0
        ):
        """ value of the agent """
        thread = CustomizedThread(target=self.query_loop, args=(query, ))
        thread.start()
        return thread.join()

    def query_loop(self, query):
        """ custom function for the CustomizedThread """
        agent_process = self.create_agent_request(query)

        # because it might have to run multiple times
        completed_response, start_times, end_times, waiting_times, turnaround_times = "", [], [], [], []

        while agent_process.get_status() != "done":
            thread = Thread(target=self.listen, args=(agent_process, ))
            current_time = time.time()
            # reinitialize agent status
            agent_process.set_created_time(current_time)
            agent_process.set_response(None)
            LLMRequestQueue.add_message(agent_process)

            thread.start()
            thread.join()

            completed_response = agent_process.get_response()
            if agent_process.get_status() != "done":
                self.logger.log(
                    f"Suspended due to the reach of time limit ({agent_process.get_time_limit()}s). Current result is: {completed_response.response_message}\n",
                    level="suspending"
                )

            # data from the AgentProcess temporarily held
            start_time = agent_process.get_start_time()
            end_time = agent_process.get_end_time()
            waiting_time = start_time - agent_process.get_created_time()
            turnaround_time = end_time - agent_process.get_created_time()

            start_times.append(start_time)
            end_times.append(end_time)
            waiting_times.append(waiting_time)
            turnaround_times.append(turnaround_time)
            # Re-start the thread if not done

        # self.agent_process_factory.deactivate_agent_process(agent_process.get_pid())

        return completed_response, start_times, end_times, waiting_times, turnaround_times

    def create_agent_request(self, query):
        """ uses AgentFactory to initialize the agent in the pool """
        agent_process = self.agent_process_factory.activate_agent_process(
            agent_name = self.agent_name,
            query = query
        )
        agent_process.set_created_time(time.time())
        # print("Already put into the queue")
        return agent_process

    def listen(self, agent_process: AgentProcess):
        """Response Listener for agent

        Args:
            agent_process (AgentProcess): Listened AgentProcess

        Returns:
            str: LLM response of Agent Process
        """
        while agent_process.get_response() is None:
            time.sleep(0.2)

        return agent_process.get_response()

    def parse_result(self, prompt):
        """ each agent has their own parser """
        pass

    #######################
    # getters and setters #
    #######################

    def set_aid(self, aid):
        self.aid = aid

    def get_aid(self):
        return self.aid

    def get_agent_name(self):
        return self.agent_name

    def set_status(self, status):

        """
        Status type: Waiting, Running, Done, Inactive
        """
        self.status = status

    def get_status(self):
        return self.status

    def set_created_time(self, time):
        self.created_time = time

    def get_created_time(self):
        return self.created_time

    def set_start_time(self, time):
        self.start_time = time

    def get_start_time(self):
        return self.start_time

    def set_end_time(self, time):
        self.end_time = time

    def get_end_time(self):
        return self.end_time



from ..base import BaseRapidAPITool

from pyopenagi.utils.utils import get_from_env

import requests

import json

class GetRestaurantDetails(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://tripadvisor16.p.rapidapi.com/api/v1/restaurant/getRestaurantDetails"
        self.host_name = "tripadvisor16.p.rapidapi.com"
        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params: dict):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }

        try:
            self.query_string = {
                "restaurantsId": params["restaurantsId"],
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for tripadvisor get restaurant details api. "
                "Please make sure it contains the key: 'restaurantsID'"
            )

        response = requests.get(self.url, headers=headers, params=self.query_string).json()
        return self.parse_result(response)


    def parse_result(self, response) -> str:
        location = response["data"]["location"]

        useful_info = {
            "name": location.get("name"),
            "latitude": location.get("latitude"),
            "longitude": location.get("longitude"),
            "num_reviews": location.get("num_reviews"),
            "rating": location.get("rating"),
            "price_level": location.get("price_level"),
            "address": location.get("address"),
            "phone": location.get("phone"),
            "website": location.get("website"),
            "cuisine": [cuisine["name"] for cuisine in location.get("cuisine", [])],
            "hours": location.get("hours", {}).get("week_ranges", [])
            }
        return json.dumps(useful_info)

    def get_tool_call_format(self):
        tool_call_format = {
			"type": "function",
			"function": {
				"name": "get_restaurant_details",
				"description": "Provides details about a restaurant",
				"parameters": {
					"type": "object",
					"properties": {
						"restaurantsId": {
							"type": "string",
							"description": "The ID of the restaurant to get details for"
						}
					},
					"required": [
						"restaurantsId"
					]
				}
			}
		}
        return tool_call_format



from ..base import BaseRapidAPITool

# from pydantic import root_validator

from pyopenagi.utils.utils import get_from_env

import requests

class SongAutoComplete(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://shazam.p.rapidapi.com/auto-complete"
        self.host_name = "shazam.p.rapidapi.com"

        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        try:
            self.query_string = {
                "term": params["term"],
                "locale": params["locale"]
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for currency converter api. "
                "Please make sure it contains two keys: 'term' and 'locale'"
            )
        response = requests.get(self.url, headers=headers, params=self.query_string).json()

        result = self.parse_result(response)
        return result


    def parse_result(self, response) -> str:
        return "Completion hints are: " + ",".join(response["hints"].values())



from ..base import BaseTool

from pyopenagi.utils.utils import get_from_env

from typing import List, Any

class GoogleSearch(BaseTool):
    """Google Search Tool, refactored from langchain.

    Adapted from: Instructions adapted from https://stackoverflow.com/questions/
    37083058/
    programmatically-searching-google-in-python-using-custom-search

    1. Install google-api-python-client
    - If you don't already have a Google account, sign up.
    - If you have never created a Google APIs Console project,
    read the Managing Projects page and create a project in the Google API Console.
    - Install the library using pip install google-api-python-client

    2. Enable the Custom Search API
    - Navigate to the APIs & Services→Dashboard panel in Cloud Console.
    - Click Enable APIs and Services.
    - Search for Custom Search API and click on it.
    - Click Enable.
    URL for it: https://console.cloud.google.com/apis/library/customsearch.googleapis
    .com

    3. To create an API key:
    - Navigate to the APIs & Services → Credentials panel in Cloud Console.
    - Select Create credentials, then select API key from the drop-down menu.
    - The API key created dialog box displays your newly created key.
    - You now have an API_KEY

    Alternatively, you can just generate an API key here:
    https://developers.google.com/custom-search/docs/paid_element#api_key

    4. Setup Custom Search Engine so you can search the entire web
    - Create a custom search engine here: https://programmablesearchengine.google.com/.
    - In `What to search` to search, pick the `Search the entire Web` option.
    After search engine is created, you can click on it and find `Search engine ID`
      on the Overview page.

    """
    def __init__(self):
        super().__init__()
        self.google_api_key = get_from_env("GOOGLE_API_KEY")
        self.google_cse_id = get_from_env("GOOGLE_CSE_ID")
        self.k: int = 10 # topk searched results
        self.search_engine = self.build_engine()
        self.siterestrict: bool = False

    def build_engine(self):
        try:
            from googleapiclient.discovery import build

        except ImportError:
            raise ImportError(
                "google-api-python-client is not installed. "
                "Please install it with `pip install google-api-python-client"
                ">=2.100.0`"
            )
        engine = build("customsearch", "v1", developerKey=self.google_api_key)
        return engine

    def _google_search_results(self, search_term: str, **kwargs: Any) -> List[dict]:
        cse = self.search_engine.cse()
        if self.siterestrict:
            cse = cse.siterestrict() # TODO add siterestrict verification
        res = cse.list(q=search_term, cx=self.google_cse_id, **kwargs).execute()
        return res.get("items", [])


    def run(self, query: str) -> str:
        """Run query through GoogleSearch and parse result."""
        response = self._google_search_results(query, num=self.k)
        result = self.parse_result(response)
        return result

    def parse_result(self, response):
        snippets = []
        if len(response) == 0:
            return "No good Google Search Result was found"
        for result in response:
            if "snippet" in result:
                snippets.append(result["snippet"])

        return " ".join(snippets)



from ..base import BaseRapidAPITool

# from pydantic import root_validator

from pyopenagi.utils.utils import get_from_env

import requests

class SongAutocompleteAPI(BaseRapidAPITool):
    def __init__(self):
        super().__init__()
        self.url = "https://ai-weather-by-meteosource.p.rapidapi.com/find_places"
        self.host_name = "ai-weather-by-meteosource.p.rapidapi.com"

        self.api_key = get_from_env("RAPID_API_KEY")

    def run(self, params):
        headers = {
            "X-RapidAPI-Key": self.api_key,
            "X-RapidAPI-Host": self.host_name
        }
        try:
            self.query_string = {
                "text": params["text"],
                "language": params["language"]
            }
        except ValueError:
            raise KeyError(
                "The keys in params do not match the excepted keys in params for weather find place api. "
                "Please make sure it contains two keys: 'text' and 'language'"
            )
        response = requests.get(self.url, headers=headers, params=self.query_string).json()

        result = self.parse_result(response)
        return result


    def parse_result(self, response) -> str:
        location = [
            response["radm_area1"],
            response["adm_area2"],
            response["country"],
            response["lat"],
            response["lon"]
        ]
        return f"Found place of {response["name"]}: " + ",".join(location)



import re

from ..base import BaseTool

import arxiv

from typing import Optional

class Arxiv(BaseTool):
    """Arxiv Tool, refactored from Langchain.

    To use, you should have the ``arxiv`` python package installed.
    https://lukasschwab.me/arxiv.py/index.html
    This wrapper will use the Arxiv API to conduct searches and
    fetch document summaries. By default, it will return the document summaries
    of the top-k results.
    If the query is in the form of arxiv identifier
    (see https://info.arxiv.org/help/find/index.html), it will return the paper
    corresponding to the arxiv identifier.
    It limits the Document content by doc_content_chars_max.
    Set doc_content_chars_max=None if you don't want to limit the content size.

    Attributes:
        top_k_results: number of the top-scored document used for the arxiv tool
        ARXIV_MAX_QUERY_LENGTH: the cut limit on the query used for the arxiv tool.
        load_max_docs: a limit to the number of loaded documents
        load_all_available_meta:
            if True: the `metadata` of the loaded Documents contains all available
            meta info (see https://lukasschwab.me/arxiv.py/index.html#Result),
            if False: the `metadata` contains only the published date, title,
            authors and summary.
        doc_content_chars_max: an optional cut limit for the length of a document's
            content
    """

    def __init__(self):
        super().__init__()
        self.top_k_results: int = 3
        self.ARXIV_MAX_QUERY_LENGTH: int = 300
        self.load_max_docs: int = 100
        self.load_all_available_meta: bool = False
        self.doc_content_chars_max: Optional[int] = 4000
        self.build_client()

    def is_arxiv_identifier(self, query: str) -> bool:
        """Check if a query is an arxiv identifier."""
        arxiv_identifier_pattern = r"\d{2}(0[1-9]|1[0-2])\.\d{4,5}(v\d+|)|\d{7}.*"
        for query_item in query[: self.ARXIV_MAX_QUERY_LENGTH].split():
            match_result = re.match(arxiv_identifier_pattern, query_item)
            if not match_result:
                return False
            assert match_result is not None
            if not match_result.group(0) == query_item:
                return False
        return True

    def build_client(self):
        """Validate that the python package exists in environment."""
        self.arxiv_search = arxiv.Search
        self.arxiv_exceptions = arxiv.ArxivError

    def run(self, params) -> str:
        """
        Performs an arxiv search and A single string
        with the publish date, title, authors, and summary
        for each article separated by two newlines.

        If an error occurs or no documents found, error text
        is returned instead. Wrapper for
        https://lukasschwab.me/arxiv.py/index.html#Search

        Args:
            query: a plaintext search query
        """  # noqa: E501
        query = params["query"]
        try:
            if self.is_arxiv_identifier(query):
                results = self.arxiv_search(
                    id_list=query.split(),
                    max_results=self.top_k_results,
                ).results()
            else:
                results = self.arxiv_search(  # type: ignore
                    query[: self.ARXIV_MAX_QUERY_LENGTH], max_results=self.top_k_results
                ).results()
        except self.arxiv_exceptions as ex:
            return f"Arxiv exception: {ex}"
        docs = [
            f"Published: {result.updated.date()}\n"
            f"Title: {result.title}\n"
            f"Authors: {', '.join(a.name for a in result.authors)}\n"
            f"Summary: {result.summary}"
            for result in results
        ]
        if docs:
            return "\n\n".join(docs)[: self.doc_content_chars_max]
        else:
            return "No good Arxiv Result was found"

    def get_tool_call_format(self):
        tool_call_format = {
            "type": "function",
            "function": {
                "name": "arxiv",
                "description": "Query articles or topics in arxiv",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "Input query that describes what to search in arxiv"
                        }
                    },
                    "required": [
                        "query"
                    ]
                }
            }
        }
        return tool_call_format




from .base_agent import BaseAgent

import time

from ..utils.chat_template import Query

import json

class ReactAgent(BaseAgent):
    def __init__(self,
                 agent_name,
                 task_input,
                 agent_process_factory,
                 log_mode: str
        ):
        BaseAgent.__init__(
            self,
            agent_name,
            task_input,
            agent_process_factory,
            log_mode
        )

        self.plan_max_fail_times = 3
        self.tool_call_max_fail_times = 3

    def build_system_instruction(self):
        prefix = "".join(
            [
                "".join(self.config["description"])
            ]
        )

        plan_instruction = "".join(
            [
                f'You are given the available tools from the tool list: {json.dumps(self.tool_info)} to help you solve problems. ',
                'Generate a plan of steps you need to take. ',
                'The plan must follow the json format as: ',
                '[',
                '{"message": "message_value1","tool_use": [tool_name1, tool_name2,...]}',
                '{"message": "message_value2", "tool_use": [tool_name1, tool_name2,...]}',
                '...',
                ']',
                'In each step of the planned workflow, you must select the most related tool to use',
                'Followings are some plan examples:',
                '[',
                '{"message": "gather information from arxiv. ", "tool_use": ["arxiv"]},',
                '{"message", "write a summarization based on the gathered information. ", "tool_use": []}',
                '];',
                '[',
                '{"message": "identify the tool that you need to call to obtain information. ", "tool_use": ["imdb_top_movies", "imdb_top_series"]},',
                '{"message", "give recommendations for the user based on the information. ", "tool_use": []}',
                '];',
            ]
        )

        if self.workflow_mode == "manual":
            self.messages.append(
                {"role": "system", "content": prefix}
            )

        else:
            assert self.workflow_mode == "automatic"
            self.messages.append(
                {"role": "system", "content": prefix + plan_instruction}
            )


    def automatic_workflow(self):
        return super().automatic_workflow()

    def manual_workflow(self):
        pass

    def call_tools(self, tool_calls):
        # self.logger.log(f"***** It starts to call external tools *****\n", level="info")
        success = True
        actions = []
        observations = []
        for tool_call in tool_calls:
            function_name = tool_call["name"]
            function_to_call = self.tool_list[function_name]
            function_params = tool_call["parameters"]

            try:
                function_response = function_to_call.run(function_params)
                actions.append(f"I will call the {function_name} with the params as {function_params}")
                observations.append(f"The knowledge I get from {function_name} is: {function_response}")

            except Exception:
                actions.append("I fail to call any tools.")
                observations.append(f"The tool parameter {function_params} is invalid.")
                success = False

        return actions, observations, success

    def run(self):
        self.build_system_instruction()

        task_input = self.task_input

        self.messages.append(
            {"role": "user", "content": task_input}
        )
        self.logger.log(f"{task_input}\n", level="info")

        workflow = None

        if self.workflow_mode == "automatic":
            workflow = self.automatic_workflow()
        else:
            assert self.workflow_mode == "manual"
            workflow = self.manual_workflow()

        self.messages.append(
            {"role": "assistant", "content": f"[Thinking]: The workflow generated for the problem is {json.dumps(workflow)}"}
        )

        self.messages.append(
            {"role": "user", "content": "[Thinking]: Follow the workflow to solve the problem step by step. "}
        )

        self.logger.log(f"Generated workflow is: {workflow}\n", level="info")

        if workflow:
            final_result = ""

            for i, step in enumerate(workflow):
                message = step["message"]
                tool_use = step["tool_use"]

                # print(f"message: {message}")
                # print(f"tool use: {tool_use}")

                prompt = f"At step {i + 1}, you need to {message}. "
                self.messages.append({
                    "role": "user",
                    "content": prompt
                })
                if tool_use:
                    selected_tools = self.pre_select_tools(tool_use)

                else:
                    selected_tools = None

                response, start_times, end_times, waiting_times, turnaround_times = self.get_response(
                    query = Query(
                        messages = self.messages,
                        tools = selected_tools
                    )
                )
                if self.rounds == 0:
                    self.set_start_time(start_times[0])

                # execute action
                response_message = response.response_message

                tool_calls = response.tool_calls

                self.request_waiting_times.extend(waiting_times)
                self.request_turnaround_times.extend(turnaround_times)

                if tool_calls:
                    for _ in range(self.plan_max_fail_times):
                        actions, observations, success = self.call_tools(tool_calls=tool_calls)

                        action_messages = "[Action]: " + ";".join(actions)
                        observation_messages = "[Observation]: " + ";".join(observations)

                        self.messages.append(
                            {
                                "role": "assistant",
                                "content": action_messages + ". " + observation_messages
                            }
                        )
                        if success:
                            break
                else:
                    thinkings = response_message
                    self.messages.append({
                        "role": "assistant",
                        "content": thinkings
                    })

                if i == len(workflow) - 1:
                    final_result = self.messages[-1]

                self.logger.log(f"At step {i + 1}, {self.messages[-1]}\n", level="info")

                self.rounds += 1

            self.set_status("done")
            self.set_end_time(time=time.time())

            return {
                "agent_name": self.agent_name,
                "result": final_result,
                "rounds": self.rounds,
                "agent_waiting_time": self.start_time - self.created_time,
                "agent_turnaround_time": self.end_time - self.created_time,
                "request_waiting_times": self.request_waiting_times,
                "request_turnaround_times": self.request_turnaround_times,
            }

        else:
            return {
                "agent_name": self.agent_name,
                "result": "Failed to generate a valid workflow in the given times.",
                "rounds": self.rounds,
                "agent_waiting_time": None,
                "agent_turnaround_time": None,
                "request_waiting_times": self.request_waiting_times,
                "request_turnaround_times": self.request_turnaround_times,
            }



# defines a process holder for agents to use in a single class
# used in the base implementation for agents
# this class isn't actually instantiated until the user instantiates
# a specific agent like MathAgent

import heapq

from threading import Thread, Lock, Event

from ..utils.chat_template import Query

class AgentProcess:
    """
    each agent holds the values defined in the constructor
    """
    def __init__(self,
            agent_name: str,
            query: Query
        ):
        """Agent Process

        Args:
            agent_name (str): Name of the agent
            query (Query): Query sent by the agent
        """
        self.agent_name = agent_name
        self.query = query
        self.pid: int = None
        self.status = None
        self.response = None
        self.time_limit = None
        self.created_time = None
        self.start_time = None
        self.end_time = None

    #######################
    # getters and setters #
    #######################

    def set_created_time(self, time):
        self.created_time = time

    def get_created_time(self):
        return self.created_time

    def set_start_time(self, time):
        self.start_time = time

    def get_start_time(self):
        return self.start_time

    def set_end_time(self, time):
        self.end_time = time

    def get_end_time(self):
        return self.end_time

    def set_priority(self, priority):
        self.priority = priority

    def get_priority(self):
        return self.priority

    def set_status(self, status):
        self.status = status

    def get_status(self):
        return self.status

    def set_pid(self, pid):
        self.pid = pid

    def get_pid(self):
        return self.pid

    def get_response(self):
        return self.response

    def set_response(self, response):
        self.response = response

    def get_time_limit(self):
        return self.time_limit

    def set_time_limit(self, time_limit):
        self.time_limit = time_limit


class LLMRequestProcess(AgentProcess):
    pass

class AgentProcessFactory:
    """ 
    used by AgentFactory
    right now it is only set in agent activation
    """

    def __init__(self, agent_process_log_mode = None):
        # make a heap of every pid 1..1024 to remove from when agent created
        # so pids can be reused
        self.max_pid = 1024
        self.pid_pool = [i for i in range(self.max_pid)]
        heapq.heapify(self.pid_pool)

        self.thread = Thread(target=self.deactivate_agent_process)

        self.current_agent_processes = dict()

        self.current_agent_processes_lock = Lock()

        self.terminate_signal = Event()

        self.agent_process_log_mode = agent_process_log_mode

    def activate_agent_process(self, agent_name, query):
        """ run each agent and lock it """
        if not self.terminate_signal.is_set():
           with self.current_agent_processes_lock:
                agent_process = AgentProcess(
                    agent_name = agent_name,
                    query = query
                )
                pid = heapq.heappop(self.pid_pool)
                agent_process.set_pid(pid)
                agent_process.set_status("active")
                self.current_agent_processes[pid] = agent_process
                return agent_process

    def print_agent_process(self):
        """ print agent data in a clean format """
        headers = ["Agent Process ID", "Agent Name", "Created Time", "Status"]
        data = []
        for id, agent_process in self.current_agent_processes.items():
            agent_name = agent_process.agent_name
            created_time = agent_process.created_time
            status = agent_process.status
            # memory_usage = f"{asizeof.asizeof(agent)} bytes"
            data.append(
                [id, agent_name, created_time, status]
            )
        self.print(headers=headers, data=data)


    def print(self, headers, data):
        """ separate headers and data in printing """

        # align output
        column_widths = [
            max(len(str(row[i])) for row in [headers] + data) for i in range(len(headers))
        ]
        print("+" + "-" * (sum(column_widths) + len(headers) * 3 - 3 ) + "+")
        print(self.format_row(headers, column_widths))
        print("=" * (sum(column_widths) + len(headers) * 3 - 1))
        for i, row in enumerate(data):
            print(self.format_row(row, column_widths))
            if i < len(data):
                print("-" * (sum(column_widths) + len(headers) * 3 - 1))
        print("+" + "-" * (sum(column_widths) + len(headers) * 3 - 3 ) + "+")


    def format_row(self, row, widths, align="<"):
        """ helper for print """
        row_str = " | ".join(f"{str(item):{align}{widths[i]}}" for i, item in enumerate(row))
        return row_str

    def deactivate_agent_process(self, pid):
        """ after agent stops reallow pid """
        self.current_agent_processes.pop(pid)
        heapq.heappush(self.pid_pool, pid)

    def start(self):
        """start the factory to check inactive agent"""
        self.thread.start()

    def stop(self):
        self.thread.join()
